from pypact.util.decorators import freeze_it
from pypact.util.numerical import is_float, get_float
from pypact.util.jsonserializable import JSONSerializable

NUCLIDE_IGNORES = ['\n', '|', '>', '&', '?', '#']


@freeze_it
class Nuclide(JSONSerializable):
    """
        The nuclide type from the output
    """
    def __init__(self):
        self.element = ""
        self.isotope = 0
        self.state = ""
        self.half_life = 0.0
        self.atoms = 0.0
        self.grams = 0.0
        self.activity = 0.0
        self.heat = 0.0
        self.alpha_heat = 0.0
        self.beta_heat = 0.0
        self.gamma_heat = 0.0
        self.dose = 0.0
        self.ingestion = 0.0
        self.inhalation = 0.0

    def fispact_deserialize(self, linedump, column_headers):

        self.__init__()

        # takes the first line of the dump and
        # strip of the markers and ignores
        line = linedump[0]
        for i in NUCLIDE_IGNORES:
            line = line.replace(i, '')

        # turn the line into a list
        strings = line.split()

        # first 2 characters are the symbol
        # next 3 characters are the symbol
        # last character is the symbol
        symbollength = 2
        isotopelength = 3
        if len(strings[0]) > symbollength:
            isotope = strings[0][symbollength:]
            strings[0] = strings[0][:symbollength]
            strings.insert(1, isotope)

        if strings[1][-1].isalpha():
            self.state = strings[1][-1]
            strings[1] = strings[1][:-1]

        self.element = strings[0]
        self.isotope = int(strings[1][:isotopelength])

        # Remove the isotope entry so that the entries now match with the header
        strings.pop(1)

        assert(len(column_headers) == len(strings))

        def get_entry(header_name):
            column_index = index_containing_substring(column_headers, header_name)
            if column_index != -1:
                item = strings[column_index]
                if is_float(item):
                    return get_float(item)

            return 0.0

        self.atoms = get_entry('ATOMS')
        self.grams = get_entry('GRAMS')
        self.activity = get_entry('Bq')
        self.half_life = get_entry('HALF LIFE')
        self.alpha_heat = get_entry('a-Energy')
        self.beta_heat = get_entry('b-Energy')
        self.gamma_heat = get_entry('g-Energy')
        # since the precision in the output file is to 4 significant figures of beta and gamma heats
        # we must set the precision to 4 significant figures for the total heat,
        # otherwise it is misleading to the user if they get 13.45999999999999999e-5 instead of 1.346e-4
        self.heat = float("{0:.4g}".format(self.alpha_heat + self.beta_heat + self.gamma_heat))
        self.dose = get_entry('DOSE RATE')
        self.ingestion = get_entry('INGESTION')
        self.inhalation = get_entry('INHALATION')


def index_containing_substring(the_list, substring):
    for i, s in enumerate(the_list):
        if substring in s:
            return i
    return -1



import os

from pypact.util.decorators import freeze_it
from pypact.util.exceptions import PypactOutOfRangeException
from pypact.util.exceptions import PypactInvalidOptionException
from pypact.util.exceptions import PypactIncompatibleOptionException
from pypact.util.jsonserializable import JSONSerializable
from pypact.util.loglevels import *
from pypact.input.keywords import CONTROL_KEYWORDS, INIT_KEYWORDS, INVENTORY_KEYWORDS
from pypact.input.projectiles import PROJECTILE_NEUTRON, VALID_PROJECTILES

COMMENT_START = '<<'
COMMENT_END   = '>>'

class InventoryType(JSONSerializable):
    def __init__(self):
        # a list of tuples, the first entry is the element/nuclide symbol
        # the second is the value
        self.entries = []

@freeze_it
class MassInventory(InventoryType):
    def __init__(self):
        super().__init__()

        # the total mass in kg
        self.totalMass = 0.0

    def __str__(self):
        strrep = "{} {} {}".format('MASS', self.totalMass, len(self.entries))
        for e in self.entries:
            strrep += "\n{} {}".format(e[0], e[1])

        return strrep

@freeze_it
class FuelInventory(InventoryType):
    def __init__(self):
        super().__init__()

    def __str__(self):
        strrep = "{} {}".format('FUEL', len(self.entries))
        for i in self.entries:
            strrep += "\n{} {}".format(i[0], i[1])

        return strrep   

@freeze_it
class InputData(JSONSerializable):

    def __init__(self, name: str = "run"):
        """
            Constructor
            
            Initialises all input data
        """
        self.name= name
        
        self._overwrite             = False
        self._json                  = False
        self._ignorecollapse        = True
        self._ignorecondense        = True
        self._condense              = False
        self._binaryxs              = False
        self._useeaf                = False
        self._approxgamma           = False
        self._outputhalflife        = False
        self._outputhazards         = False
        self._initialinventory      = False
        self._readgammagroup        = False
        self._readspontfission      = False
        self._ignoreuncert          = False
        self._enablemonitor         = False
        self._usecumfissyield       = False
        self._clearancedata         = False
        self._loglevel              = LOG_SEVERITY_WARNING
        
        # default is 1.0E-12 barns
        self._xsthreshold           = 1.0e-12
        self._group                 = 0
        self._projectile            = PROJECTILE_NEUTRON

        # set the minimum number of atoms deemed significant for the inventory output
        self._atomsthreshold       = 0.0
        
        # the total mass in grams per cubic centimetre (g/cc)
        self._density              = 0.0

        self._inventoryismass      = False
        self._inventoryisfuel      = False 
        
        self._inventorymass        = MassInventory()
        self._inventoryfuel        = FuelInventory()
    
        # irradiation schedule
        # a list of tuples of (time interval in seconds, flux amplitude)
        self._irradschedule = []
        
        # cooling schedule
        # a list of time interval in seconds
        self._coolingschedule = []
    
    def reset(self):
        """
            Reset the data
            
            Reinitialises all input data to defaults
        """
        self.__init__(self.name)

    def validate(self):
        """
            Validate the input data
            
            Not implemented yet
        """
        # to do
        pass

    def overwriteExisting(self, overwrite = True):
        """
            Enables overwriting of output files
            uses keyword CLOBBER
        """
        self._overwrite = overwrite

    def enableJSON(self, enable = True):
        """
            Enables JSON output file
            uses keyword JSON
        """
        self._json = enable
    
    def enableInitialInventoryInOutput(self, output = True):
        """
            Performs the time = 0, inventory step,
            the initial inventory
        """
        self._initialinventory = output
    
    def enableHalflifeInOutput(self, output = True):
        """
            Enables half lives to be written to the output file
        """
        self._outputhalflife = output
    
    def enableHazardsInOutput(self, output = True):
        self._outputhazards = output
    
    def readXSData(self, group, binary = False):
        if self._useeaf and binary:
            raise PypactIncompatibleOptionException("Cannot enable binary format reading when using EAF")

        self._ignorecollapse = False
        self._group = group
        self._binaryxs = binary
    
    def useEAFLibraries(self, use = True):
        if self._binaryxs and use:
            raise PypactIncompatibleOptionException("Cannot enable EAF data libraries with binary format")

        self._useeaf = use
    
    def useCumulativeFissionYieldData(self, use = True):
        self._usecumfissyield = use
    
    def includeClearanceData(self, include = True):
        self._clearancedata = include
    
    def readDecayData(self, condense = True):
        self._ignorecondense = False
        self._condense = condense
    
    def approxGammaSpectrum(self, approxgamma = True):
        self._approxgamma = approxgamma
    
    def ignoreUncertainties(self, ignore = True):
        self._ignoreuncert = ignore
    
    def setXSThreshold(self, threshold):
        self._xsthreshold = threshold
    
    def setProjectile(self, proj):
        if proj not in VALID_PROJECTILES:
            raise PypactInvalidOptionException("{} is not a valid projectile option.")
            
        self._projectile = proj

    def readGammaGroup(self, readgg = True):
        self._readgammagroup = readgg
    
    def enableSystemMonitor(self, enable = True):
        self._enablemonitor = enable

    def setAtomsThreshold(self, threshold):
        self._atomsthreshold = threshold
    
    def addIrradiation(self, timeInSecs, fluxAmp):
        self._irradschedule.append((timeInSecs, fluxAmp))
    
    def resetIrradiation(self):
        self._irradschedule = []
    
    def addCooling(self, timeInSecs):
        self._coolingschedule.append(timeInSecs)
    
    def resetCooling(self):
        self._coolingschedule = []
    
    def setLogLevel(self, severity):
        if severity < LOG_SEVERITY_FATAL or severity > LOG_SEVERITY_TRACE:
            raise PypactOutOfRangeException("Log level {} not valid.".format(severity))
        self._loglevel = severity

    def setDensity(self, densityInGPCC):
        """
            Sets the density of the target in g/cc
            densityInGPCC: density in g/cc, must be positive
        """
        if not densityInGPCC > 0.0:
            raise PypactOutOfRangeException("Density must be positive.")

        self._density = densityInGPCC

    def setMass(self, totalMassInKg):
        """
            Sets the material mode to Mass
            Sets the total mass of the target in kg
            totalMassInKg: mass in kg, must be positive
        """
        if not totalMassInKg > 0.0:
            raise PypactOutOfRangeException("Total mass must be positive.")

        self._inventorymass.totalMass = totalMassInKg
        self._inventoryismass = True
        self._inventoryisfuel = False

    def setFuel(self):
        """
            Sets the material mode to Fuel
        """

        self._inventoryismass = False
        self._inventoryisfuel = True

    def addIsotope(self, isotope: str, numberOfAtoms):
        """
            Add an isotope
            isotope: character symbol of element name, e.g. 'Fe' and the mass number of the isotpe e.g '56'
            numberOfAtoms: the number of atoms present 
        """
        if numberOfAtoms < 0:
            raise PypactUnphysicalValueException("Number of atoms must be positive")
        #could check for integer value and raise here PypactTypeException
        
        self._inventoryfuel.entries.append((isotope, numberOfAtoms))

    def clearIsotopes(self):
        self._inventoryfuel.entries = []

    def addElement(self, element, percentage = 100.0):
        """
            Add an element
            element: character symbol of element name, e.g. 'Fe'
            percentage: percentage contribution of total mass, should not exceed 100%
        """
        if percentage > 100.0:
            raise PypactOutOfRangeException("Cannot set the element percentage above 100%.")
        
        self._inventorymass.entries.append((element, percentage))
        
    def clearElements(self):
        self._inventorymass.entries = []
    
    def _serialize(self, f):
        """
            The serialization method
            f: file object
        """
        inputdata = []
        
        def addnewline():
            inputdata.append("")
        
        def addcomment(comment):
            inputdata.append("{} {} {}".format(COMMENT_START, comment, COMMENT_END))
        
        def addkeyword(keyword, args=[]):
            strargs = ' '.join([str(a) for a in args])
            inputdata.append("{} {}".format(keyword, strargs))
        
        # control keywords
        addcomment("CONTROL PHASE")
        if self._json:
            addcomment("enable JSON output")
            addkeyword('JSON')
                
        if self._overwrite:
            addcomment("overwrite existing output files of same name")
            addkeyword('CLOBBER')
        
        if self._readgammagroup:
            addcomment("read gamma groups from file, specify ggbins in files file")
            addkeyword('READGG')
        
        if self._readspontfission:
            addcomment("read spontaneous fission from file, specify sf_endf in files file")
            addkeyword('READSF')
        
        if self._useeaf:
            addcomment("use EAF nuclear data libraries")
            addkeyword('LIBVERSION', args=[0])
        
        if self._usecumfissyield:
            addcomment("use cumulative fission yield data mt=459 instead of mt=454")
            addkeyword('CUMFYLD')
            
        if self._enablemonitor:
            addcomment("monitor FISPACT-II progress")
            addkeyword('MONITOR', args=[1])
        
        addcomment("the minimum cross section (barns) for inclusion in pathways analysis")
        addkeyword('XSTHRESHOLD', args=[self._xsthreshold])
        
        if not self._ignorecollapse:
            if self._group != 0:
                addcomment("perform collapse")
                addkeyword('GETXS', args=[-1 if self._binaryxs and not self._useeaf else 1, self._group])
            else:
                addcomment("don't do collapse, just read the existing file")
                addkeyword('GETXS', args=[0])

        if not self._ignorecondense:
            addcomment("get decay data")
            addkeyword('GETDECAY', args=[1 if self._condense else 0])
    
        if self._loglevel != LOG_SEVERITY_WARNING:
            addcomment("enable logging at level {}".format(self._loglevel))
            addkeyword('LOGLEVEL', args=[self._loglevel])

        if self._approxgamma:
            addcomment("approximate spectra when not available")
            addkeyword('SPEK')
        
        if self._ignoreuncert:
            addcomment("ignore uncertainties")
            addkeyword('NOERROR')
    
        if self._projectile != PROJECTILE_NEUTRON:
            addcomment("set projectile (n=1, d=2, p=3, a=4, g=5)")
            addkeyword('PROJ', args=[self._projectile])
        
        # end control phase
        addcomment("end control")
        addkeyword('FISPACT')
        addkeyword('*', args=[self.name])

        # initial phase
        addnewline()
        addcomment("INITIALIZATION PHASE")
        if self._outputhalflife:
            addcomment("output half life values")
            addkeyword('HALF')

        if self._outputhazards:
            addcomment("output ingestion and inhalation values")
            addkeyword('HAZARDS')

        if self._clearancedata:
            addcomment("include clearance data of radionuclides to be input")
            addkeyword('CLEAR')
            
        if self._inventoryismass and not self._inventoryisfuel:
            if self._inventorymass.totalMass > 0.0:
                addcomment("set the target via MASS")
                addkeyword(str(self._inventorymass))

        if self._inventoryisfuel and not self._inventoryismass:
            if self._density > 0.0:
                addcomment("set the target via FUEL")
                addkeyword(str(self._inventoryfuel))  
        
        if self._density > 0.0:
            addcomment("set the target density")
            addkeyword('DENSITY', args=[self._density])
        
        if self._atomsthreshold > 0.0:
            addcomment("set the threshold for atoms in the inventory")
            addkeyword('MIND', args=[self._atomsthreshold])
        
        if self._initialinventory:
            addcomment("output the initial inventory")
            addkeyword('ATOMS')

        # inventory phase
        addnewline()
        addcomment("INVENTORY PHASE")
        if len(self._irradschedule) > 0:
            addcomment("irradiation schedule")
            for time, fluxamp in self._irradschedule:
                addkeyword('FLUX', args=[fluxamp])
                addkeyword('TIME', args=[time, 'SECS'])
                addkeyword('ATOMS')
            addcomment("end of irradiation")

            addkeyword('FLUX', args=[0.0])
            addkeyword('ZERO')
            for time in self._coolingschedule:
                addkeyword('TIME', args=[time, 'SECS'])
                addkeyword('ATOMS')
            addcomment("end of cooling")

        # end file
        addnewline()
        addcomment("end of input")
        addkeyword('END')
        addkeyword('*', args=['end'])
        
        for line in inputdata:
            f.write("{}\n".format(line))

    def _deserialize(self, f):
        """
            The deserialization method
            f: file object
        """
        pass





import os
from pypact.input.fluxesfile import FluxesFile
from pypact.input.filesfile import FilesFile
from pypact.input.inputdata import InputData
from pypact.util.decorators import freeze_it
from pypact.util.jsonserializable import JSONSerializable


@freeze_it
class FispactInput(JSONSerializable):
    """
        Full FISPACT-II super input class
        that contains everything needed to run FISPACT-II.
        
        Contains the input data (input file), flux data (fluxes file) and
        nuclear data and file information (files file).
        
        This is then JSON serializable which can be used for a RESTful API.
    """
    def __init__(self, runname="run", fluxname="flux", nuclear_data_base=os.sep):
        self.input = InputData(name=runname)
        self.flux  = FluxesFile(name=fluxname)
        self.files = FilesFile(base_dir=nuclear_data_base)



from pypact.util.decorators import freeze_it
from pypact.util.lines import first_occurrence
from pypact.util.jsonserializable import JSONSerializable
from pypact.output.nuclide import Nuclide
from pypact.output.tags import NUCLIDES_HEADER
import pypact.util.propertyfinder as pf

NUCLIDES_IGNORES = ['\n', '|']


@freeze_it
class Nuclides(JSONSerializable):
    """
        The nuclides type from the output
    """
    def __init__(self):
        self.nuclides = []

    def __len__(self):
        return len(self.nuclides)

    def __getitem__(self, index):
        return self.nuclides[index]

    def json_deserialize(self, j, objtype=object):
        super(Nuclides, self).json_deserialize(j)
        self.json_deserialize_list(j, 'nuclides', Nuclide)

    def fispact_deserialize(self, filerecord, interval):

        self.__init__()

        substring = filerecord[interval]

        nuclidetag = 'TOTAL NUMBER OF NUCLIDES PRINTED IN INVENTORY'

        def nr_of_nuclides():
            number = pf.first(datadump=substring,
                              headertag=NUCLIDES_HEADER,
                              starttag=nuclidetag,
                              endtag='',
                              ignores=NUCLIDES_IGNORES,
                              asstring=False)
            return int(number)

        # The column headers line is after the main nuclide header
        def get_header(index):
            raw = substring[index].split('  ')
            header = list(filter(''.__ne__, raw))
            for ignore in NUCLIDES_IGNORES:
                if ignore in header:
                    header = list(filter(ignore.__ne__, header))

            return header

        # The nuclides list starts from the line prior to the total number
        # so we need the line number of this tag and count backwards
        i, line = first_occurrence(lines=substring, tag=nuclidetag)
        if i < 0:
            return

        for n in range(i - nr_of_nuclides(), i):
            nuclide = Nuclide()
            nuclide.fispact_deserialize(substring[n:],
                                        column_headers=get_header(i - nr_of_nuclides()-3))
            self.nuclides.append(nuclide)



from pypact.util.decorators import freeze_it
from pypact.util.numerical import get_float
from pypact.util.lines import first_value_from_line, strings_from_line
from pypact.util.jsonserializable import JSONSerializable
from pypact.output.tags import DOSE_RATE_HEADER
import pypact.util.propertyfinder as pf

DOSE_RATE_IGNORES = ['\n', '|']


@freeze_it
class DoseRate(JSONSerializable):
    """
        The dose rate type from the output
    """
    def __init__(self):
        self.type = ""
        self.distance = 0.0
        self.mass = 0.0
        self.dose = 0.0

    def fispact_deserialize(self, filerecord, interval):

        self.__init__()

        substring = filerecord[interval]

        dose_string = pf.first(datadump=substring,
                               headertag=DOSE_RATE_HEADER,
                               starttag=DOSE_RATE_HEADER,
                               endtag=')',
                               ignores=DOSE_RATE_IGNORES,
                               asstring=True)

        if 'PLANE' in dose_string:
            self.type = 'PLANE SOURCE'
        elif 'POINT' in dose_string:
            self.type = 'POINT SOURCE'

            # Mass is always in grams and is actually always 1 g
            # but read it in to verify
            self.mass = first_value_from_line(dose_string, '')*1e-3

            # Distance is always in meters but is written without a space so
            # we must strip it off
            floatstr = strings_from_line(dose_string, 'SOURCE')[-1].replace('m', '')
            self.distance = get_float(floatstr)

        self.dose = pf.first(datadump=substring,
                             headertag=DOSE_RATE_HEADER,
                             starttag='IS',
                             endtag='Sieverts/hour',
                             ignores=DOSE_RATE_IGNORES,
                             asstring=False)



# output readers
from pypact.reader import Reader
from pypact.filerecord import FileRecord
from pypact.runner import compute

# output data structures
from pypact.output.rundata import RunData
from pypact.output.doserate import DoseRate
from pypact.output.nuclide import Nuclide
from pypact.output.nuclides import Nuclides
from pypact.output.output import Output
from pypact.output.gammaspectrum import GammaSpectrum
from pypact.output.timestep import TimeStep

# input
from pypact.input.fispactinput import FispactInput
from pypact.input.fluxesfile import FluxesFile, ArbFluxesFile
from pypact.input.filesfile import FilesFile
from pypact.input.inputdata import InputData
from pypact.input.serialization import to_file, from_file, to_string
from pypact.input.groupstructures import ALL_GROUPS
from pypact.input.keywords import CONTROL_KEYWORDS
from pypact.input.keywords import INIT_KEYWORDS
from pypact.input.keywords import INVENTORY_KEYWORDS
from pypact.input.keywords import OVER_SUBKEYWORDS
from pypact.input.keywords import DEPRECATED_KEYWORDS
from pypact.input.projectiles import PROJECTILE_NEUTRON
from pypact.input.projectiles import PROJECTILE_DEUTERON
from pypact.input.projectiles import PROJECTILE_PROTON
from pypact.input.projectiles import PROJECTILE_ALPHA
from pypact.input.projectiles import PROJECTILE_GAMMA
from pypact.input.projectiles import get_projectile_name
from pypact.input.projectiles import get_projectile_symbol
from pypact.input.projectiles import get_projectile_value
from pypact.input.projectiles import VALID_PROJECTILE_NAMES
from pypact.input.projectiles import VALID_PROJECTILE_SYMBOLS
from pypact.input.projectiles import VALID_PROJECTILES

# library
from pypact.library.nuclidelib import NUCLIDE_DICTIONARY
from pypact.library.nuclidelib import NUMBER_OF_ELEMENTS
from pypact.library.nuclidelib import NUMBER_OF_ISOTOPES
from pypact.library.nuclidelib import find_isotopes
from pypact.library.nuclidelib import find_element
from pypact.library.nuclidelib import find_z
from pypact.library.nuclidelib import get_all_isotopes

# utilities
from pypact.util.exceptions import PypactException
from pypact.util.loglevels import *

# This makes importing slow, keep it seperate
#import pypact.analysis



#!/usr/bin/env python3
import sys
import argparse

from pypact.reader import Reader
from pypact.util.exceptions import PypactException


def main():
    # Command line argument support
    parser = argparse.ArgumentParser(description='Fispact Output Converter')
    parser.add_argument('outputfile', type=argparse.FileType('r'),
                        help='The fispact output (.out) file to be read')
    parser.add_argument('jsonoutputfile', type=argparse.FileType('w'),
                        help='The fispact output file in JSON format to be written')
    args = parser.parse_args()

    filename = args.outputfile.name

    try:
        with Reader(filename) as output:
            with open(args.jsonoutputfile.name, 'w') as jsonfile:
                jsonfile.write(output.json_serialize())

    except PypactException as err:
        print("Pypact error: {0}".format(err))
    except OSError as err:
        print("OS error: {0}".format(err))
    except ValueError as err:
        print(err)
    except:
        print("Unexpected error:", sys.exc_info()[0])


if __name__ == "__main__":
    main()



import os.path


# Returns the content of the file as a string
def content_as_str(filename):
    with open(filename, 'r') as file:
        data = file.readlines()

    return data


# Returns the content of the file as a string, and removes new lines.
def content_as_str_noreturn(filename):
    with open(filename, 'r') as file:
        data = file.read().replace('\n', '')

    return data


# checks if file exists
def file_exists(filename):
    return os.path.isfile(filename)


# checks if directory exists
def dir_exists(filename):
    return os.path.isdir(filename)


def file_remove(filename):
    if file_exists(filename):
        os.remove(filename)


def get_filename_ext(filename):
    return os.path.basename(os.path.splitext(filename)[-1])


def str_in_file(filename, string):
    if file_exists(filename):
        with open(filename) as file:
            return string in file.read()

    return False


# note that regardless of the flag 'ignore_empty_lines' value,
# the last line is always ignored
def nr_of_lines(filename, ignore_empty_lines=False):
    with open(filename) as file:
        if ignore_empty_lines:
            return len([i for i in file if i[:-1]])

        return len([i for i in file])



from itertools import accumulate
from pypact.util.file import content_as_str
from pypact.util.lines import line_indices
from pypact.output.tags import TIME_STEP_HEADER, IRRAD_TIME_TAG, COOLING_TIME_TAG
import pypact.util.propertyfinder as pf


class FileRecord:

    def __init__(self, filename, asstring=''):
        """
        Cache the file content as a list of strings and process the timesteps
        :param filename:
        """

        self.cachedlines = (asstring if asstring else content_as_str(filename))
        self.lineindices = line_indices(self.cachedlines, TIME_STEP_HEADER)
        self.timesteps = []
        self.irradiation_times = []
        self.cooling_times = []

        self._processtimesteps()

    def __len__(self):
        return len(self.lineindices)

    def __getitem__(self, interval):
        """
        Get the timestep lines for a given interval

        :param interval:
        :return:
        """
        l = [s for i, s in self.timesteps if i == interval]
        if len(l) == 1:
            return l[0]

        return ''

    def cumulirradiationtime(self, interval):
        """
        Get the cumulative irradiation time at a given interval

        :param interval:
        :return:
        """
        return self._getcumultime(interval, self.irradiation_times)

    def cumulcoolingtime(self, interval):
        """
        Get the cumulative irradiation time at a given interval

        :param interval:
        :return:
        """
        return self._getcumultime(interval, self.cooling_times)

    def _processtimesteps(self):

        for i in range(0, len(self)):
            t = self.lineindices[i]
            nt = -1
            if i < len(self.lineindices) - 1:
                nt = self.lineindices[i+1]

            interval = int(pf.first(datadump=self.cachedlines[t:nt],
                                    headertag=TIME_STEP_HEADER,
                                    starttag=TIME_STEP_HEADER,
                                    endtag='',
                                    ignores=[],
                                    asstring=False))

            irrad_time = pf.first(datadump=self.cachedlines[t:nt],
                                  headertag=TIME_STEP_HEADER,
                                  starttag=IRRAD_TIME_TAG,
                                  endtag='SECS',
                                  ignores=[],
                                  asstring=False)

            cool_time = pf.first(datadump=self.cachedlines[t:nt],
                                 headertag=TIME_STEP_HEADER,
                                 starttag=COOLING_TIME_TAG,
                                 endtag='SECS',
                                 ignores=[],
                                 asstring=False)

            self.timesteps.append((interval, self.cachedlines[t:nt]))
            self.irradiation_times.append(irrad_time)
            self.cooling_times.append(cool_time)

        # turn them into cumulative values
        self.irradiation_times = list(accumulate(self.irradiation_times))
        self.cooling_times = list(accumulate(self.cooling_times))

        assert len(self.irradiation_times) == len(self.cooling_times)

    def _getcumultime(self, interval, listoftuples):
        t = [x for x, y in enumerate(self.timesteps) if y[0] == interval]
        if len(t) == 1:
            return listoftuples[t[0]]

        return 0.0



from pypact.util.numerical import is_float, get_float


def line_indices(lines, tag):
    """

    :param lines: Must be a list of strings representing each line
    :param tag: is the tag to find in the list of lines
    :return: the indices of the lines with that tag

    """
    return [l for l in range(len(lines)) if tag in lines[l]]


def strings_from_line(line, linetag, ignoretags=[], endtag=''):
    """

    :param line: The string representing a line
    :param linetag: The tag to find in the line
    :param ignoretags: The list of strings to ignore
    :param endtag: The tag of which to stop the string
    :return: A list of strings, without spaces or carriage returns, of string after the linetag

    """
    startindex = line.find(linetag)
    if startindex < 0:
        return []

    endindex = len(line)
    if endtag:
        endindex = line.find(endtag)
        if endindex < 0:
            endindex = len(line)

    strings = (line[startindex + len(linetag):endindex]).split()

    # trim
    for tag in ignoretags:
        if tag in strings:
            strings.remove(tag)

    return strings


def join_strings_from_line(line, linetag, ignoretags=[], endtag=''):
    """

    :param line: The string representing a line
    :param linetag: The tag to find in the line
    :param ignoretags: The list of strings to ignore
    :param endtag: The tag of which to stop the string
    :return: A single string joining all strings from the list

    """
    strings = strings_from_line(line, linetag, ignoretags, endtag)

    return ' '.join(strings)


def first_value_from_line(line, linetag, ignoretags=[]):
    """

    :param line: The string representing a line
    :param linetag: The tag to find in the line
    :param ignoretags: The list of strings to ignore
    :return: The first value, float, after the tag. Returns NaN if no value found.

    """
    strings = strings_from_line(line, linetag, ignoretags)

    nan = 0.0
    if not strings:
        return nan

    for s in strings:
        # strip off any common tags around numbers
        s = s.replace(',', '').replace('*', '')
        if is_float(s):
            return get_float(s)

    return nan


def first_occurrence(lines, tag):
    """

    :param lines: A list of strings representing each line
    :param tag: is the tag to find in the list of lines
    :return: A tuple of the index and the line

    """
    return next_occurrence(lines, tag, 0)


def last_occurrence(lines, tag):
    """

    :param lines: A list of strings representing each line
    :param tag: is the tag to find in the list of lines
    :return: A tuple of the index and the line

    """
    lineindices = line_indices(lines, tag)
    if len(lineindices) > 0:
        index = lineindices[-1]
        line = lines[index]
        return index, line.strip()
    else:
        return -1, ''


def next_occurrence(lines, tag, startindex=0):
    """

    :param lines: A list of strings representing each line
    :param tag: is the tag to find in the list of lines
    :return: A tuple of the index and the line

    """
    lineindices = line_indices(lines[startindex:], tag)
    if len(lineindices) > 0:
        index = lineindices[0]+startindex
        line = lines[index]
        return index, line.strip()
    else:
        return -1, ''




import os

from pypact.filerecord import FileRecord
from pypact.output.output import Output


class Reader:
    def __init__(self, filename, ignorenuclides=False):
        """
            It can read both JSON and out file formats
        """
        self.record = FileRecord(filename)
        self.output = Output(ignorenuclides=ignorenuclides)
        self.isjson = False
        if os.path.splitext(filename)[-1] == '.json':
            self.isjson = True
            with open(filename) as f:
                self.output.json_deserialize(f.read())

    def __enter__(self):
        if not self.isjson:
            self.output.fispact_deserialize(self.record)

        return self.output

    def __exit__(self, *args):
        pass



nan = float('NaN')

def are_values_the_same(value1, value2, rel_tol, abs_tol=0.0):
    """
        Use math.isclose algorithm, it is better than numpy.isclose method.
        See https://github.com/numpy/numpy/issues/10161 for more on the discussion
        Since some python version don't come with math.isclose we implement it here directly
    """
    return abs(value1 - value2) <= max(rel_tol * max(abs(value1), abs(value2)), abs_tol)


def get_float(value):
    """
        Gets the floating point value from a string
        Will allow for fortran style floats, i.e -2.34321-308
        If it is neither then it will return "nan"
    """
    if _istradiationalfloat(value):
        return float(value)

    return _getfortranfloat(value)


def _istradiationalfloat(value):
    """
        Checks if the string can be converted to a floating point value
        Does not allow for fortran style floats, i.e -2.34321-308
        only standard floats.
    """
    try:
        float(value)
        return True
    except ValueError:
        return False

def is_float(value):
    """
        Checks if the string can be converted to a floating point value
        Will allow for fortran style floats, i.e -2.34321-308
        If it is neither then it will return False
    """
    if isinstance(value, (int, float, str)):
        return _istradiationalfloat(value) or _isfortranfloat(value)
    return False

def _isfortranfloat(value):
    passfunc = lambda sign, esign, parts: True
    failfunc = lambda: False

    return _fortranfloat(value, passfunc, failfunc)


def _getfortranfloat(value):
    passfunc = lambda sign, esign, parts: float(sign + parts[0] + 'E' + esign + parts[1])
    failfunc = lambda: "nan"

    return _fortranfloat(value, passfunc, failfunc)


def _fortranfloat(value, passfunc, failfunc):
    # could be 2.3
    #          2.3e+10
    #          -2.3e+10
    #          -2.3+10
    #          -2.3-10
    #          2.3-10
    #          +2.3-10
    #          +2.3+10
    #          -2.3+10

    signs = ['-', '+']

    valueasstring = str(value)
    sign = ""
    if valueasstring:
        # check for sign at the front
        if valueasstring[0] in signs:
            sign = valueasstring[0]
            valueasstring = valueasstring[1:]

        # cannot have both separators in the value for a valid float
        if not all(s in valueasstring for s in signs):
            # check the values in both parts recursively
            for sn in signs:
                if sn in valueasstring:
                    parts = valueasstring.split(sn, 1)
                    if _istradiationalfloat(parts[0]) and \
                       _istradiationalfloat(parts[1]) and \
                       '.' in parts[0] and not '.' in parts[1]:
                        return passfunc(sign, sn, parts)
    return failfunc()



from pypact.analysis.plotadapter import LinePlotAdapter
from pypact.analysis.plotadapter import MatrixPlotAdapter
from pypact.analysis.plotadapter import AnimatedMatrixPlotAdapter

from pypact.analysis.propertyplotter import NuclideDataEntry
from pypact.analysis.propertyplotter import plotproperty

from pypact.analysis.timezone import TimeZone

from pypact.library.nuclidelib import find_isotopes
from pypact.library.nuclidelib import find_element
from pypact.library.nuclidelib import find_z
from pypact.library.nuclidelib import get_all_isotopes



import os

from pypact.util.decorators import freeze_it
from pypact.util.jsonserializable import JSONSerializable
from pypact.util.exceptions import PypactDeserializeException
from pypact.util.file import file_exists, dir_exists
from pypact.input.projectiles import *

# represents a null entry string
NULL_ENTRY = "null"

NUCLEAR_LIBS = {
            'TENDL2014':
            {
                'ind_nuc' : os.path.join('TENDL2014data', 'tendl14_decay12_index'),
                'xs_endf' : os.path.join('TENDL2014data', 'tal2014-n', 'gxs-709'),
                'prob_tab' : os.path.join('TENDL2014data', 'tal2014-n', 'tp-709-294'),
            },
            'TENDL2015':
            {
                'ind_nuc' : os.path.join('TENDL2015data', 'tendl15_decay12_index'),
                'xs_endf' : os.path.join('TENDL2015data', 'tal2015-n', 'gxs-709'),
                'prob_tab' : os.path.join('TENDL2015data', 'tal2015-n', 'tp-709-294')
            },
            'TENDL2017':
            {
                'ind_nuc' : os.path.join('TENDL2017data', 'tendl17_decay12_index'),
                'xs_endf' : os.path.join('TENDL2017data', 'tal2017-n', 'gxs-709'),
                'prob_tab' : os.path.join('TENDL2017data', 'tal2017-n', 'tp-709-294')
            },
            'EAF2010':
            {
                'ind_nuc' : os.path.join('EAF2010data', 'eaf_index_20100'),
                'crossec' : os.path.join('EAF2010data', 'eaf_n_gxs_066_fis_20100'),
                'crossunc' : os.path.join('EAF2010data', 'eaf_un_20100'),
                'decay' : os.path.join('EAF2010data', 'eaf_dec_20100.001'),
                'asscfy' : os.path.join('EAF2010data', 'eaf_n_asscfy_20100'),
                'fissyld' : os.path.join('EAF2010data', 'eaf_n_fis_20100'),
                'hazards' : os.path.join('EAF2010data', 'eaf_haz_20100'),
                'clear' : os.path.join('EAF2010data', 'eaf_clear_20100'),
                'a2data' : os.path.join('EAF2010data', 'eaf_a2_20100'),
                'absorp' : os.path.join('EAF2010data', 'eaf_abs_20100')
            },
            'CENDL31':
            {
                'ind_nuc' : os.path.join('CENDL31data', 'cendl31_decay12_index'),
                'xs_endf' : os.path.join('CENDL31data', 'cendl31-n', 'gxs-709'),
            },
            'ENDFB71':
            {
                'ind_nuc' : os.path.join('ENDFB71data', 'endfb71_index'),
                'xs_endf' : os.path.join('ENDFB71data', 'endfb71-n', 'gxs-709'),
                'dk_endf' : os.path.join('ENDFB71data', 'decay'),
                'fy_endf' : os.path.join('ENDFB71data', 'endfb71nfy'),
                'sf_endf' : os.path.join('ENDFB71data', 'endfb71sfy')
            },
            'ENDFB80':
            {
                'ind_nuc' : os.path.join('ENDFB80data', 'endfb80_index'),
                'xs_endf' : os.path.join('ENDFB80data', 'endfb80-n', 'gxs-709'),
                'dk_endf' : os.path.join('ENDFB80data', 'decay')
            },
            'GEFY42':
            {
                'fy_endf' : os.path.join('GEFY42data', 'gefy42_nfy'),
                'sf_endf' : os.path.join('GEFY42data', 'gefy42_sfy')
            },
            'GEFY52':
            {
                'fy_endf' : os.path.join('GEFY52data', 'gef52_nfy'),
                'sf_endf' : os.path.join('GEFY52data', 'gef52_sfy')
            },
            'GEFY61':
            {
                'fy_endf' : os.path.join('GEFY61data', 'gefy61_nfy'),
                'sf_endf' : os.path.join('GEFY61data', 'gefy61_sfy')
            },
            'UKFY41':
            {
                'fy_endf' : os.path.join('UKFY41data', 'ukfy4_1p'),
            },
            'UKFY42':
            {
                'fy_endf' : os.path.join('UKFY42data', 'ukfy4_2n'),
            },
            'HEAD2009':
            {
                'ind_nuc' : os.path.join('HEAD2009data', 'head2009-n_decay12_index'),
                'xs_endf' : os.path.join('ENDFB71data', 'head2009-n', 'gxs-sp-709'),
            },
            'HEIR01':
            {
                'ind_nuc' : os.path.join('HEIR01data', 'heir01_decay12_index'),
                'xs_endf' : os.path.join('HEIR01data', 'heir01-p', 'gxs-162'),
            },
            'JEFF32':
            {
                'ind_nuc' : os.path.join('JEFF32data', 'jeff32_decay12_index'),
                'xs_endf' : os.path.join('JEFF32data', 'jeff32-n', 'gxs-709'),
                'dk_endf' : os.path.join('JEFF32data', 'decay'),
                'fy_endf' : os.path.join('JEFF32data', 'jeff32-n', 'jeff311nfy'),
                'sf_endf' : os.path.join('JEFF32data', 'jeff32-n', 'jeff311sfy')
            },
            'JEFF33':
            {
                'ind_nuc' : os.path.join('JEFF33data', 'jeff33_decay12_index'),
                'xs_endf' : os.path.join('JEFF33data', 'jeff33-n', 'gxs-709'),
                'prob_tab' : os.path.join('JEFF33data', 'jeff33-n', 'tp-709-294'),
                'dk_endf' : os.path.join('JEFF33data', 'decay'),
                'fy_endf' : os.path.join('JEFF33data', 'jeff33-n', 'jeff33nfy'),
                'sf_endf' : os.path.join('JEFF33data', 'jeff33-n', 'jeff33sfy')
            },
            'JENDL4data':
            {
                'ind_nuc' : os.path.join('JENDL4data', 'jendl4_decay12_index'),
                'xs_endf' : os.path.join('JENDL4data', 'jendl4-n', 'gxs-709'),
                'dk_endf' : os.path.join('JENDL4data', 'decay'),
                'fy_endf' : os.path.join('JENDL4data', 'jendl4-n', 'jendl4nfy'),
                'sf_endf' : os.path.join('JENDL4data', 'jendl4-n', 'jendl4sfy')
            },
            'DECAY':
            {
                'dk_endf' : os.path.join('decay', 'decay_2012'),
                'hazards' : os.path.join('decay', 'hazards_2012'),
                'clear'   : os.path.join('decay', 'clear_2012'),
                'a2data'  : os.path.join('decay', 'a2_2012'),
                'absorp'  : os.path.join('decay', 'abs_2012'),
            },
}

@freeze_it
class FilesFile(JSONSerializable):
    def __init__(self, base_dir=os.sep, group=709, projectile=PROJECTILE_NEUTRON):
        self.reset()

        self.__base_dir = base_dir
        self.__group = group
        self.__projectile = projectile

        # defaults
        self.fluxes     = "fluxes"
        self.collapxi   = "COLLAPX"
        self.collapxo   = "COLLAPX"
        self.arrayx     = "ARRAYX"

    def reset(self):
        # nuclear data libraries
        self.ind_nuc    = NULL_ENTRY
        self.xs_endf    = NULL_ENTRY
        self.xs_endfb   = NULL_ENTRY
        self.prob_tab   = NULL_ENTRY
        self.fy_endf    = NULL_ENTRY
        self.asscfy     = NULL_ENTRY
        self.dk_endf    = NULL_ENTRY
        self.hazards    = NULL_ENTRY
        self.clear      = NULL_ENTRY
        self.a2data     = NULL_ENTRY
        self.absorp     = NULL_ENTRY
        self.sf_endf    = NULL_ENTRY
        self.sp_endf    = NULL_ENTRY
        self.ggbins     = NULL_ENTRY
        self.enbins     = NULL_ENTRY
        self.xs_extra   = NULL_ENTRY
        
        # eaf nuclear data
        self.crossec    = NULL_ENTRY
        self.crossunc   = NULL_ENTRY
        self.decay      = NULL_ENTRY
        self.fissyld    = NULL_ENTRY
        
        # additional files  - not nuclear data
        self.fluxes     = NULL_ENTRY
        self.arb_flux   = NULL_ENTRY
        self.collapxi   = NULL_ENTRY
        self.collapxo   = NULL_ENTRY
        self.arrayx     = NULL_ENTRY
        self.files      = NULL_ENTRY
        self.input      = NULL_ENTRY
        self.graph      = NULL_ENTRY
        self.gnuplot    = NULL_ENTRY
        self.spec       = NULL_ENTRY
        self.output     = NULL_ENTRY
        self.tab1       = NULL_ENTRY
        self.tab2       = NULL_ENTRY
        self.tab3       = NULL_ENTRY
        self.tab4       = NULL_ENTRY
        self.runlog     = NULL_ENTRY
        self.ind_nuco   = NULL_ENTRY
        self.sens       = NULL_ENTRY
        self.nucgraph   = NULL_ENTRY
        self.nucgnu     = NULL_ENTRY
    
    def to_dict(self):
        d = dict()
        for k, v in self.__dict__.items():
            if '__' not in k and v != NULL_ENTRY:
                d[k] = v

        return d

    def setXS(self, type):
        self._setVar(type, 'ind_nuc')
        if type == 'EAF2010':
            self._setVar(type, 'crossec')
            self._setVar(type, 'crossunc')
        else:
            self._setVar(type, 'xs_endf')

    def setProbTab(self, type):
        self._setVar(type, 'prob_tab')

    def setFissionYield(self, type):
        self._setVar(type, 'fy_endf')

    def setDecay(self, type):
        key = 'dk_endf'
        if type == 'EAF2010':
            key = 'decay'

        self._setVar(type, key)

    def setEAF(self):
        self.setXS('EAF2010')
        self.setDecay('EAF2010')
        self.setRegulatory('EAF2010')
        self.setGammaAbsorb('EAF2010')
    
    def setRegulatory(self, type):
        self._setVar(type, 'hazards')
        self._setVar(type, 'clear')
        self._setVar(type, 'a2data')

    def setGammaAbsorb(self, type):
        self._setVar(type, 'absorp')

    def _setVar(self, type, key):
        if type not in NUCLEAR_LIBS and \
            not key in NUCLEAR_LIBS[type]:
            raise PypactException(
                    "Cannot set {} library to type {}, does not exist.".format(key, type))

        self.__setattr__(key, os.path.join(self.__base_dir, NUCLEAR_LIBS[type][key]))

    def invalidpaths(self):
        """
            Validate if paths in files file exist
            Return a list of tuples (key, value) that do not exist
        """
        ignore = ['collapxi',
                  'collapxo',
                  'arrayx',
                  'tab1',
                  'tab2',
                  'tab3',
                  'tab4',
                  'output',
                  'ind_nuco']
        invalid = []
        for k, v in self.to_dict().items():
            if k in ignore:
                continue
            
            if not v or (not file_exists(v) and not dir_exists(v)):
                invalid.append((k,v))

        return invalid

    def _serialize(self, f):
        """
            The serialization method
            f: file object
            
            Format is:
            
            key1          value1
            
            key2          value2
            
            ...
        """
        for k, v in self.to_dict().items():
            f.write("{:<20} {:<100}\n\n".format(k,v))

    def _deserialize(self, f):
        """
            The deserialization method
            f: file object
            
            Format is:
            
            key1          value1
            
            key2          value2
        """
        for line in f:
            data = line.split()
            if not data:
                continue
            
            k = data[0]

            # ignore comments
            if k[0] == '#':
                continue

            if k not in self.__dict__:
                raise PypactDeserializeException(
                    "Cannot deserialize files file. Unknown key {0} specified.".format(k))

            if self.__dict__[k] != NULL_ENTRY:
                raise PypactDeserializeException(
                    "Cannot deserialize files file. Repeated key {0} specified.".format(k))
            
            self.__setattr__(k, data[1])




from __future__ import division

from pypact.analysis.timezone import TimeZone


class NuclideDataEntry(object):
    def __init__(self, nuclide):
        self.element = nuclide[0]
        self.isotope = nuclide[1]

        self.reset()

    def reset(self):
        self.times = []
        self.values = []

    def __str__(self):
        return str.format("Isotope {0}-{1} values: {2}",
                          self.element, self.isotope, self.values)


def plotproperty(output,
                 property,
                 isotopes,
                 plotter,
                 fractional=False,
                 timeperiod=TimeZone.BOTH):

    plotter.newcanvas()
    [i.reset() for i in isotopes]

    # process the data
    for t in output:
        if (t.cooling_time > 0.0 and timeperiod == TimeZone.IRRAD) or \
        (t.cooling_time == 0.0 and timeperiod == TimeZone.COOL):
            continue

        total = 0.0
        for n in t.nuclides:
            value = getattr(n, property)
            total += value
            for i in isotopes:
                if n.element == i.element and n.isotope == i.isotope:
                    i.times.append(t.irradiation_time + t.cooling_time)
                    i.values.append(value)

        if fractional and total > 0:
            for i in isotopes:
                for v in i.values:
                    v = v / total

    # style the plot
    for i in isotopes:
        if not i.values or not all(j > 0 for j in i.values):
            continue

        yaxislabel = str.format('{0}', property)
        if fractional:
            yaxislabel = str.format('fractional {0}', property)

        f = plotter.lineplot(x=i.times,
                            y=i.values,
                            datalabel=str.format('{0}-{1}', i.element, i.isotope),
                            xlabel="time [s]",
                            ylabel=yaxislabel,
                            logx=(timeperiod != TimeZone.IRRAD),
                            logy=True,
                            overlay=True)

    if timeperiod == TimeZone.BOTH:
        plotter.custom('axvline', x=output[-1].irradiation_time,
                      color='k', linestyle='--', label='cooling')
    plotter.addlegend(location='lower right')

    return f



import json


class JSONSerializable(object):

    def json_serialize(self):
        def obj_dict(obj):
            new_dict = {}
            for k, v in obj.__dict__.items():
                if '__' not in k:
                    new_dict[k] = v
            return new_dict

        return json.dumps(self, default=obj_dict, indent=4, sort_keys=False)

    def json_deserialize(self, j, objtype=object):
        """
        Read the JSON file into the JSONSerializable data objects

        :param j: The JSON dump
        :param key: The key/member variable name
        :param objtype: The type of object to be added to the list
        :return:
        """
        d = json.loads(j)
        for key, value in self.__dict__.items():
            if key in d:
                if isinstance(value, JSONSerializable):
                    value.json_deserialize(json.dumps(d[key]))
                elif isinstance(d[key], dict):
                    self.json_deserialize(json.dumps(d[key]))
                elif isinstance(value, list):
                    self.json_deserialize_list(json.dumps(d[key]), key, objtype=objtype)
                else:
                    setattr(self, key, d[key])

    def json_deserialize_list(self, j, key, objtype):
        """
        This is a little hack, but I cannot think of a better way to deserialize a
        list of objects

        :param j: The JSON dump
        :param key: The key/member variable name
        :param objtype: The time of object to be added to the list
        :return:
        """
        d = json.loads(j)

        def do_append(a):
            # for the standard case of a list of str, int or floats
            if isinstance(a, (int, float, str)):
                self.__dict__[key].append(a)
                return

            # for the unique case of a list of JSONSerializable objects
            t = objtype()
            if isinstance(t, JSONSerializable):
                t.json_deserialize(json.dumps(a))
                self.__dict__[key].append(t)

        if key in d:
            for i in d[key]:
                if isinstance(self.__dict__[key], list):
                    do_append(i)
        else:
            if isinstance(d, list):
                for l in d:
                    do_append(l)



class PypactException(Exception):
    pass

class PypactFrozenException(PypactException):
    pass

class PypactInvalidOptionException(PypactException):
    pass

class PypactIncompatibleOptionException(PypactException):
    pass

class PypactOutOfRangeException(PypactException):
    pass

class PypactSerializeException(PypactException):
    pass

class PypactDeserializeException(PypactException):
    pass

class PypactFispactExecutableNotFoundException(PypactException):
    pass

class PypactTypeException(PypactException):
    pass

class PypactUnphysicalValueException(PypactException):
    pass



from pypact.util.decorators import freeze_it
from pypact.util.jsonserializable import JSONSerializable
from pypact.output.doserate import DoseRate
from pypact.output.nuclides import Nuclides
from pypact.output.gammaspectrum import GammaSpectrum
from pypact.output.tags import TIME_STEP_HEADER
import pypact.util.propertyfinder as pf

TIME_STEP_IGNORES = []


@freeze_it
class TimeStep(JSONSerializable):
    """
        An object to represent a time step in the output
    """
    def __init__(self, ignorenuclides=False):
        self.irradiation_time = 0.0
        self.cooling_time = 0.0
        self.flux = 0.0
        self.total_heat = 0.0
        self.alpha_heat = 0.0
        self.beta_heat = 0.0
        self.gamma_heat = 0.0
        self.ingestion_dose = 0.0
        self.inhalation_dose = 0.0
        self.total_activity = 0.0
        self.total_activity_exclude_trit = 0.0
        self.initial_mass = 0.0
        self.total_mass = 0.0
        self.number_of_fissions = 0.0
        self.burnup = 0.0
        self.gamma_spectrum = GammaSpectrum()
        self.dose_rate = DoseRate()
        self.nuclides = Nuclides()

        self.__ignorenuclides = ignorenuclides

    @property
    def isirradiation(self):
        return self.cooling_time == 0.0

    @property
    def currenttime(self):
        if self.isirradiation:
            return self.irradiation_time
        return self.cooling_time

    def fispact_deserialize(self, filerecord, interval):

        # reset to defaults before reading
        self.__init__(ignorenuclides=self.__ignorenuclides)

        self.irradiation_time = filerecord.cumulirradiationtime(interval)
        self.cooling_time = filerecord.cumulcoolingtime(interval)

        substring = filerecord[interval]

        def get_value(starttag, endtag):
            return pf.first(datadump=substring,
                            headertag=TIME_STEP_HEADER,
                            starttag=starttag,
                            endtag=endtag,
                            ignores=TIME_STEP_IGNORES,
                            asstring=False)

        self.flux = get_value(starttag='* * * FLUX AMP IS', endtag='/cm^2/s')

        self.alpha_heat = get_value(starttag='TOTAL ALPHA HEAT PRODUCTION', endtag='kW')
        self.beta_heat = get_value(starttag='TOTAL BETA  HEAT PRODUCTION', endtag='kW')
        self.gamma_heat = get_value(starttag='TOTAL GAMMA HEAT PRODUCTION', endtag='kW')
        self.total_heat = self.alpha_heat + self.beta_heat + self.gamma_heat

        self.initial_mass = get_value(starttag='INITIAL TOTAL MASS OF MATERIAL', endtag='kg')
        self.total_mass = get_value(starttag='TOTAL MASS OF MATERIAL', endtag='kg')
        
        self.number_of_fissions = get_value(starttag='NUMBER OF FISSIONS', endtag='BURN-UP')
        self.burnup = get_value(starttag='BURN-UP OF ACTINIDES', endtag='%')

        self.ingestion_dose = get_value(starttag='INGESTION  HAZARD FOR ALL MATERIALS', endtag='Sv/kg')
        self.inhalation_dose = get_value(starttag='INHALATION HAZARD FOR ALL MATERIALS', endtag='Sv/kg')

        self.total_activity = get_value(starttag='TOTAL ACTIVITY FOR ALL MATERIALS', endtag='Bq')
        self.total_activity_exclude_trit = get_value(starttag='TOTAL ACTIVITY EXCLUDING TRITIUM', endtag='Bq')

        self.dose_rate.fispact_deserialize(filerecord, interval)
        self.gamma_spectrum.fispact_deserialize(filerecord, interval)

        # for fission this can be very large and hence very slow
        # if you do not care about the nuclides then we can set
        # it to ignore nuclides
        if not self.__ignorenuclides:
            self.nuclides.fispact_deserialize(filerecord, interval)



import time
from functools import wraps

from pypact.util.exceptions import PypactFrozenException


def freeze_it(cls):
    cls.__frozen = False

    def frozensetattr(self, key, value):
        if self.__frozen and not hasattr(self, key):
            raise PypactException("Class {} is frozen. Cannot set {} = {}"
                  .format(cls.__name__, key, value))
        else:
            object.__setattr__(self, key, value)

    def init_decorator(func):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            func(self, *args, **kwargs)
            self.__frozen = True
        return wrapper

    cls.__setattr__ = frozensetattr
    cls.__init__ = init_decorator(cls.__init__)

    return cls

def time_it(method):
    def timed(*args, **kwargs):
        ts = time.time()
        result = method(*args, **kwargs)
        te = time.time()
        
        if 'runtime' in kwargs:
            name = kwargs.get('runname', method.__name__.upper())
            kwargs['runtime'][name] = int((te - ts))
            kwargs['runtimeunit'][name] = "seconds"
        else:
            print(' *** Time taken for {0}:  {1:.4g} seconds'.format(method.__name__, (te - ts)))
        return result
    return timed



TIME_STEP_HEADER = '* * * TIME INTERVAL'
IRRAD_TIME_TAG = '* TIME IS'
COOLING_TIME_TAG = '* COOLING TIME IS'

DOSE_RATE_HEADER = 'DOSE RATE ('

NUCLIDE_HEADER = TIME_STEP_HEADER
NUCLIDES_HEADER = NUCLIDE_HEADER

GAMMA_SPECTRUM_HEADER = 'GAMMA SPECTRUM AND ENERGIES/SECOND'
GAMMA_SPECTRUM_SUB_HEADER = 'GAMMA RAY POWER FROM ACTIVATION DECAY'



import os

from pypact.input.groupstructures import ALL_GROUPS
from pypact.util.decorators import freeze_it
from pypact.util.jsonserializable import JSONSerializable
from pypact.util.numerical import get_float, is_float
from pypact.util.exceptions import PypactException, PypactOutOfRangeException, PypactDeserializeException


@freeze_it
class FluxesFile(JSONSerializable):
    """
        FISPACT-II traditionally takes the fluxes file in descending order
        not ascending, hence the reason for the reversed.
        However the values and __boundaries arrays should be in ascending order.
        It is advised that the user should set the values using the setValue
        at a specific energy.
    """
    def __init__(self, name="fluxes", norm=1.0):
        self.name = name
        self.norm = norm
    
        self.__boundaries = []
        self.__midpointenergies = []
        self.values = []

    def __len__(self):
        return len(self.values)

    def reset(self):
        self.__init__(name=self.name)   

    @property
    def boundaries(self):
        return self.__boundaries
    
    @property
    def midPointEnergies(self):
        return self.__midpointenergies
    
    def setGroup(self, group):
        if group not in ALL_GROUPS:
            raise PypactOutOfRangeException("Group {} is not a valid group".format(group))

        # group structures are in reverse order
        self._setBoundaries(group)
        self.values = [0.0]*group

    def setValue(self, energy, value):
        """
            Requires setGroup is set before
        """
        if not self.__boundaries:
            raise PypactException("No group set, cannot set value.")
        
        if value < 0.0:
            raise PypactOutOfRangeException("Flux value cannot be negative.")
        
        if energy < self.__boundaries[0]:
            raise PypactOutOfRangeException("Energy value below minimum for group, group min is {}.".format(self.__boundaries[0]))
        
        if energy >= self.__boundaries[-1]:
            raise PypactOutOfRangeException("Energy value exceeds maximum for group, group max is {}.".format(self.__boundaries[-1]))
        
        for i in range(0, len(self.__boundaries)-1):
            if self.__boundaries[i+1] > energy:
                self.values[i] = value
                return

    def validate(self):
        if len(self.__boundaries) != len(self.values) + 1:
            raise PypactOutOfRangeException("Bin boundaries must be of size one greater than values size")

    def _setFromReversedBoundaries(self, boundaries):
        self.__boundaries = list(reversed(boundaries))
        self.__midpointenergies = [(self.__boundaries[i] + self.__boundaries[i+1])/2.0 for i in range(0, len(boundaries)-1)]

    def _setBoundaries(self, group):
        self._setFromReversedBoundaries(ALL_GROUPS[group])
    
    def _serialize(self, f):
        """
            The serialization method
            f: file object
            
            NOTE: The values are in reverse order
            
            Format is:
            
            v1
            v2
            ...
            vn
            1.0
            name
        """
        for e in list(reversed(self.values)):
            f.write("{}\n".format(e))
        f.write("{}\n".format(self.norm))
        f.write(self.name)

    def _deserialize(self, f):
        """
            The deserialization method
            f: file object
            
            NOTE: The values are in reverse order
            
            Format is:
            
            v1
            v2
            ...
            vn
            1.0
            name
        """
        self.reset()
        
        lines = f.readlines()
        # last two lines are the normalisation and the name
        self.name = str(lines[-1])
        self.norm = get_float(lines[-2])

        reversed_values = []
        for l in lines[:-2]:
            for e in l.split():
                if is_float(e):
                    reversed_values.append(get_float(e))
                else:
                    raise PypactDeserializeException("Entry {} in line {} is not a float.".format(e, l))

        group = len(reversed_values)
        if group not in ALL_GROUPS:
            raise PypactDeserializeException("Group structure {} not known to pypact.".format(group))

        self.values = list(reversed(reversed_values))
        self._setFromReversedBoundaries(ALL_GROUPS[group])

@freeze_it
class ArbFluxesFile(FluxesFile):

    def __init__(self, *args, **kwargs):
        super(ArbFluxesFile, self).__init__(*args, **kwargs)

    # custom boundaries should be in normal order
    def setGroup(self, boundaries):
        self._setBoundaries(boundaries)
        self.values = [0.0]*(len(boundaries)-1)
    
    # custom boundaries should be in normal order
    def _setBoundaries(self, boundaries):
        self._setFromReversedBoundaries(list(reversed(boundaries)))
        
    def _serialize(self, f):
        """
            The serialization method
            f: file object
            
            Format is:
            
            e1
            e2
            ...
            en+1
            
            v1
            v2
            ...
            vn
            1.0
            name
        """
        for e in list(reversed(self.boundaries)):
            f.write("{}\n".format(e))
        
        # write space between xs boundaries and values
        f.write("\n")
        
        for e in list(reversed(self.values)):
            f.write("{}\n".format(e))
        f.write("{}\n".format(self.norm))
        f.write(self.name)

    def _deserialize(self, f):
        """
            The deserialization method
            f: file object
            
            Format is:
            
            e1
            e2
            ...
            en+1
            
            v1
            v2
            ...
            vn
            1.0
            name
        """
        self.reset()
        
        lines = f.readlines()
        # last two lines are the normalisation and the name
        self.name = str(lines[-1])
        self.norm =  get_float(lines[-2])

        found_values = False
        reversed_values = []
        reversed_bounds = []
        for l in lines[:-2]:
            # find the space seperator to indicate flux values not boundaries
            if not l.split():
                found_values = True
            for e in l.split():
                if is_float(e):
                    if found_values:
                        reversed_values.append(get_float(e))
                    else:
                        reversed_bounds.append(get_float(e))
                else:
                    raise PypactDeserializeException("Entry {} in line {} is not a float.".format(e, l))

        self.values = list(reversed(reversed_values))
        self._setFromReversedBoundaries(reversed_bounds)



#!/usr/bin/env python3
import sys
import argparse
import os

from pypact.util.exceptions import PypactException
from pypact.input.filesfile import FilesFile
from pypact.input.serialization import serialize, deserialize


def main():
    # Command line argument support
    parser = argparse.ArgumentParser(description='Files File Maker')
    parser.add_argument('filesfile', type=argparse.FileType('w'),
                        help='The fispact files file to create (files)')
    parser.add_argument("-d", "--basedir", type=str,
                        help='The base directory')
    args = parser.parse_args()

    if args.basedir is None:
        args.basedir = os.sep
    
    filename = args.filesfile.name

    try:
        ff = FilesFile(base_dir=args.basedir)

        # set paths here, example only now
        ff.setXS('TENDL2017')
        ff.setFissionYield('GEFY52')
        ff.setProbTab('TENDL2015')
        ff.setDecay('JEFF33')
        ff.setRegulatory('DECAY')
        ff.setGammaAbsorb('DECAY')
        
        invalids = ff.validate()
        if invalids:
            print("Warning: Invalid paths in files file!")
            for i in invalids:
                print(" *** Missing {}: {}".format(i[0], i[1]))

        # write to file
        serialize(ff, filename)
    except PypactException as err:
        print("Pypact error: {}".format(err))
    except OSError as err:
        print("OS error: {}".format(err))
    except ValueError as err:
        print(err)
    except:
        print("Unexpected error:", sys.exc_info()[1])


if __name__ == "__main__":
    main()



from pypact.util.decorators import freeze_it
from pypact.util.exceptions import PypactDeserializeException
from pypact.util.lines import first_value_from_line, strings_from_line
from pypact.util.jsonserializable import JSONSerializable
from pypact.output.tags import DOSE_RATE_HEADER
import pypact.util.propertyfinder as pf
from pypact.util.lines import first_occurrence, first_value_from_line

GAMMA_SPECTRUM_IGNORES = ['\n']


@freeze_it
class GammaSpectrum(JSONSerializable):
    """
        The gamma spectrum type from the output
    """
    def __init__(self):
        self.boundaries = []
        self.values     = []

    def fispact_deserialize(self, filerecord, interval):
        self.__init__()

        substring = filerecord[interval]

        # get the
        startindex, startline = first_occurrence(lines=substring, tag='GAMMA RAY POWER FROM ACTIVATION DECAY  MeV/s    (')
        endindex, endline = first_occurrence(lines=substring, tag=DOSE_RATE_HEADER)
        if startindex < 0 or endindex < 0:
            return
        
        end = endindex-startindex-1
        for i in range(0, end):
            line = substring[startindex+i:startindex+i+1][0]
            # we need to do this since in the output file the lower bin has a - without a whitespace after the value
            boundaries_line = line.split('-')
            if len(boundaries_line) < 2:
                raise PypactDeserializeException("Error when reading gamma spectrum, expected format as ( lb- ub MeV).")
            
            lower_bin = first_value_from_line(boundaries_line[0], '(', ignoretags=[])
            self.boundaries.append(lower_bin)
            
            # since we should have n+1 boundaries for n bin values we must append the final upper bound at the end
            if i == end-1:
                upper_bin = first_value_from_line(boundaries_line[1], '', ignoretags=[])
                self.boundaries.append(upper_bin)
            
            # the bin value
            value = first_value_from_line(line, 'MeV)', ignoretags=[])
            self.values.append(value)



# control keywords
MAX_CONTROL_KEYWORDS = 32
CONTROL_KEYWORDS = [
    'ALLDISPEN',
    'ATDISPEN',
    'CLOBBER',
    'CNVTYPE',
    'COVARIANCE',
    'CUMFYLD',
    'FISPACT',
    'FULLXS',
    'GETDECAY',
    'GETXS',
    'GRPCONVERT',
    'JSON',
    'LIBVERSION',
    'LIMGRP',
    'LOGLEVEL',
    'MONITOR',
    'NOERROR',
    'NOFISS',
    'PROBTABLE',
    'PROJECTILE',
    'READGG',
    'READSF',
    'SAVELINES',
    'SPEK',
    'SSFCHOOSE',
    'SSFDILUTION',
    'SSFFUEL',
    'SSFGEOMETRY',
    'SSFMASS',
    'USESPALLATION',
    'USEXSEXTRA',
    'XSTHRESHOLD'
]
assert len(CONTROL_KEYWORDS) == MAX_CONTROL_KEYWORDS


# initialisation keywords
MAX_INIT_KEYWORDS = 64
INIT_KEYWORDS = [
    'ATOMS',
    'ATWO',
    'BREMSSTRAHLUNG',
    'CLEAR',
    'CULTAB',
    'DENSITY',
    'DEPLETION',
    'DOSE',
    'END',
    'ERROR',
    'FISCHOOSE',
    'FISYIELD',
    'FLUX',
    'FUEL',
    'FULLXS',
    'GENERIC',
    'GETXS',
    'GRAPH',
    'GROUP',
    'GRPCONVERT',
    'HALF',
    'HAZARDS',
    'INDEXPATH',
    'IRON',
    'LIBVERSION',
    'LOGLEVEL',
    'LOOKAHEAD',
    'MASS',
    'MCSAMPLE',
    'MCSEED',
    'MIND',
    'NOCOMP',
    'NOSORT',
    'NOT1',
    'NOT2',
    'NOT3',
    'NOT4',
    'NUCGRAPH',
    'OVER',
    'PATH',
    'PATHRESET',
    'POWER',
    'PRINTLIB',
    'PROBTABLE',
    'ROUTES',
    'SENSITIVITY',
    'SORTDOMINANT',
    'SPECTRUM',
    'SPLIT',
    'SSFCHOOSE',
    'SSFDILUTION',
    'SSFFUEL',
    'SSFGEOMETRY',
    'SSFMASS',
    'TAB1',
    'TAB2',
    'TAB3',
    'TAB4',
    'TIME',
    'TOLERANCE',
    'UNCERTAINTY',
    'UNCTYPE',
    'USEFISSION',
    'WALL',
]
assert len(INIT_KEYWORDS) == MAX_INIT_KEYWORDS

# inventory keywords
MAX_INVENTORY_KEYWORDS = 41
INVENTORY_KEYWORDS = [
    'ATOMS',
    'DEPLETION',
    'DAYS',
    'END',
    'ENDPULSE',
    'FLUX',
    'FULLXS',
    'GETXS',
    'GRPCONVERT',
    'HOURS',
    'LIBVERSION',
    'LOGLEVEL',
    'MINS',
    'NOSTABLE',
    'NOT1',
    'NOT2',
    'NOT3',
    'NOT4',
    'OVER',
    'PARTITION',
    'PATHRESET',
    'POWER',
    'PROBTABLE',
    'PULSE',
    'RESULT',
    'SECS',
    'SPECTRUM',
    'SSFCHOOSE',
    'SSFDILUTION',
    'SSFFUEL',
    'SSFGEOMETRY',
    'SSFMASS',
    'STEP',
    'TAB1',
    'TAB2',
    'TAB3',
    'TAB4',
    'TIME',
    'WALL',
    'YEARS',
    'ZERO',
]
assert len(INVENTORY_KEYWORDS) == MAX_INVENTORY_KEYWORDS

# over sub-keywords
MAX_OVER_SUBKEYWORDS = 4
OVER_SUBKEYWORDS = [
    'ACROSS',
    'ADCROSS',
    'ADLAM',
    'ALAM',
]
assert len(OVER_SUBKEYWORDS) == MAX_OVER_SUBKEYWORDS

# all FISPACT keywords
MAX_DEPRECATED_KEYWORDS = 15
DEPRECATED_KEYWORDS = [
    'AINPUT',       # deprecated - replaced by GETXS 0 GETDECAY 0
    'ARRAY',        # deprecated - replaced by GETDECAY 0
    'COLLAPSE',     # deprecated - replaced by GETXS 1
    'CONV',         # deprecated - use TOLERANCE instead
    'DOMINANT',     # deprecated - use UNCERTAINTY and SORTDOMINANT instead
    'EAFVERSION',   # deprecated - does nothing, use LIBVERSION instead
    'ENFA',         # deprecated - replaces GETDECAY
    'LEVEL',        # deprecated - does nothing
    'LINA',         # deprecated - replaces 0 for GETDECAY
    'LOOPS',        # deprecated - does nothing
    'NEWFILE',      # deprecated - does nothing
    'NOHEADER',     # deprecated - does nothing
    'SEQNUMBER',    # deprecated - does nothing, SCPR not yet implemented
    'SEQUENTIAL',   # deprecated - does nothing, SCPR not yet implemented
    'TAPA',         # deprecated - replaces 1 for GETDECAY
]
assert len(DEPRECATED_KEYWORDS) == MAX_DEPRECATED_KEYWORDS

# all keywords (no duplicates)
# expected number of keywords
MAX_KEYWORDS = 116
ALL_KEYWORDS = sorted(list(set(CONTROL_KEYWORDS +
                               INIT_KEYWORDS +
                               INVENTORY_KEYWORDS +
                               OVER_SUBKEYWORDS +
                               DEPRECATED_KEYWORDS)))
assert len(ALL_KEYWORDS) == MAX_KEYWORDS




# fispact values for projectiles
# do not change this mapping
PROJECTILE_NEUTRON  = 1
PROJECTILE_DEUTERON = 2
PROJECTILE_PROTON  = 3
PROJECTILE_ALPHA    = 4
PROJECTILE_GAMMA    = 5

VALID_PROJECTILES = [
    PROJECTILE_NEUTRON,
    PROJECTILE_DEUTERON,
    PROJECTILE_PROTON,
    PROJECTILE_ALPHA,
    PROJECTILE_GAMMA
]

VALID_PROJECTILE_NAMES = [
    'neutron',
    'deuteron',
    'proton',
    'alpha',
    'gamma'
]

VALID_PROJECTILE_SYMBOLS = [
    'n',
    'd',
    'p',
    'a',
    'g'
]

def get_projectile_name(projectile_value):
    return VALID_PROJECTILE_NAMES[VALID_PROJECTILES.index(projectile_value)]

def get_projectile_symbol(projectile_value):
    return VALID_PROJECTILE_SYMBOLS[VALID_PROJECTILES.index(projectile_value)]

def get_projectile_value(projectile_name):
    return VALID_PROJECTILES[VALID_PROJECTILE_NAMES.index(projectile_name)]



from pypact.util.decorators import freeze_it
from pypact.util.jsonserializable import JSONSerializable
import pypact.util.propertyfinder as pf

RUN_DATA_IGNORES = ['\n', '|']


@freeze_it
class RunData(JSONSerializable):
    """
        The run data type from the output
    """
    def __init__(self):
        self.timestamp = ""
        self.run_name = ""
        self.flux_name = ""

    def fispact_deserialize(self, filerecord):
        self.__init__()

        self.timestamp = pf.first(datadump=filerecord.cachedlines,
                                  headertag='THIS RUN',
                                  starttag='timestamp:',
                                  endtag='',
                                  ignores=RUN_DATA_IGNORES,
                                  asstring=True)

        self.run_name = pf.first(datadump=filerecord.cachedlines,
                                 headertag='THIS RUN',
                                 starttag='FISPACT title:',
                                 endtag='',
                                 ignores=RUN_DATA_IGNORES,
                                 asstring=True)

        self.flux_name = pf.first(datadump=filerecord.cachedlines,
                                  headertag='INITIAL CROSS SECTION DATA',
                                  starttag='FLUX file label:',
                                  endtag='',
                                  ignores=RUN_DATA_IGNORES,
                                  asstring=True)



from setuptools import setup


setup(name='pypact',
      version='1.3.1',
      description='The package for reading and manipulating the fispact output text file.',
      url='https://github.com/fispact/pypact',
      author='UKAEA',
      author_email='thomas.stainer@ukaea.uk',
      license='Apache License 2.0',
      packages=[
            'pypact',
            'pypact.analysis',
            'pypact.library',
            'pypact.input',
            'pypact.output',
            'pypact.tools',
            'pypact.util'
      ],
      install_requires=[],
      python_requires='>=3',
      scripts=['pypact/tools/fispactconverter.py'],
      setup_requires=['pytest-runner'],
      test_suite='tests.testsuite',
      tests_require=['pytest', 'mock'],
      zip_safe=False)



from pypact.util.decorators import freeze_it
from pypact.util.jsonserializable import JSONSerializable
from pypact.output.rundata import RunData
from pypact.output.timestep import TimeStep


@freeze_it
class Output(JSONSerializable):
    """
        An object to represent the output
    """
    def __init__(self, ignorenuclides=False):
        self.run_data = RunData()
        self.inventory_data = []

        self.__ignorenuclides = ignorenuclides

    def __len__(self):
        return len(self.inventory_data)

    def __getitem__(self, index):
        return self.inventory_data[index]

    def json_deserialize(self, j, objtype=object):
        super(Output, self).json_deserialize(j)
        self.json_deserialize_list(j, 'inventory_data', TimeStep)

    def fispact_deserialize(self, filerecord):

        self.__init__(ignorenuclides=self.__ignorenuclides)

        self.run_data.fispact_deserialize(filerecord)

        for i, s in filerecord.timesteps:
            t = TimeStep(ignorenuclides=self.__ignorenuclides)
            t.fispact_deserialize(filerecord, interval=i)
            self.inventory_data.append(t)



from pypact.util.lines import *


def get(datadump, headertag, starttag, endtag='', findfirst=True, ignores=[], asstring=True):
    """
    Get the first occurrence of the linetag within the first headertag.

    Parameters
    ----------
    datadump : str
        Full output string dump
    headertag : str
        The string tag to indicate the property is below this header tag
    starttag : str
        The string tag to indicate the property is after this tag on the line
    endtag : str
        The string tag to indicate the property is before this tag on the line (optional)
    findfirst : bool
        The flag to indicate if the first or last occurrence should be used
    ignores : list[str]
        List of strings to ignore when performing the join (optional)
    asstring : bool
        True if return as a string, False if you require the first value -
            float or int from that line after the tag (optional)
    """
    # find the first or last occurrence of the header tag in the output dump
    i, line = (first_occurrence(datadump, headertag) if findfirst
               else last_occurrence(datadump, headertag))

    # if it cannot find anything then return an empty string
    if i == -1:
        return ('' if asstring else 0.0)

    # get the first line occurrence after the header
    i, line = first_occurrence(datadump[i:], starttag)

    return (join_strings_from_line(line, starttag, ignores, endtag) if asstring
            else first_value_from_line(line, starttag, ignores))
    

def first(datadump, headertag, starttag, endtag='', ignores=[], asstring=True):
    """
    Get the first occurrence of the linetag within the first headertag.

    Parameters
    ----------
    datadump : str
        Full output string dump
    headertag : str
        The string tag to indicate the property is below this header tag
    starttag : str
        The string tag to indicate the property is after this tag on the line
    endtag : str
        The string tag to indicate the property is before this tag on the line (optional)
    ignores : list[str]
        List of strings to ignore when performing the join (optional)
    asstring : bool
        True if return as a string, False if you require the first value -
            float or int from that line after the tag (optional)
    """
    return get(datadump=datadump,
               headertag=headertag,
               starttag=starttag,
               endtag=endtag,
               findfirst=True,
               ignores=ignores,
               asstring=asstring)


def last(datadump, headertag, starttag, endtag='', ignores=[], asstring=True):
    """
    Get the last occurrence of the linetag within the last headertag.

    Parameters
    ----------
    datadump : str
        Full output string dump
    headertag : str
        The string tag to indicate the property is below this header tag
    starttag : str
        The string tag to indicate the property is after this tag on the line
    endtag : str
        The string tag to indicate the property is before this tag on the line (optional)
    ignores : list[str]
        List of strings to ignore when performing the join (optional)
    asstring : bool
        True if return as a string, False if you require the first value -
            float or int from that line after the tag (optional)
    """
    return get(datadump=datadump,
               headertag=headertag,
               starttag=starttag,
               endtag=endtag,
               findfirst=False,
               ignores=ignores,
               asstring=asstring)



try:
    import matplotlib.pyplot
    from matplotlib import cm
    from matplotlib.colors import LogNorm
    import matplotlib.animation as animation
    from matplotlib.patches import Rectangle
    from matplotlib.collections import PatchCollection
except:
    raise ImportError("Matplotlib cannot be found. It is required for plotting.")

try:
    import numpy as np
except:
    raise ImportError("Numpy cannot be found. It is required for plotting.")


class PlotAdapter(object):
    """
        Wraps the Matplotlib plotter
    """
    _engine_name = 'matplotlib'
    _engine = matplotlib.pyplot

    def __repr__(self):
        return "<%s: %s>" % (self.__class__.__name__, self.enginename)

    @property
    def enginename(self):
        """
        The name of the plotting engine
        """
        return self._engine_name

    @property
    def engine(self):
        """
        The plotter engine
        """
        return self._engine

    def show(self):
        self.engine.show()

    def grid(self, show=True):
        self.engine.grid(show)

    def addlegend(self, location):
        self.engine.legend(loc=location)

    def newcanvas(self, *args, **kwargs):
        self._figure = self.engine.figure(*args, **kwargs)
        return self._figure


class LinePlotAdapter(PlotAdapter):
    def lineplot(self, x, y, datalabel="", xlabel="", ylabel="",
                 logx=False, logy=False, overlay=True):

        if not overlay:
            self.newcanvas()

        self.engine.xlabel(xlabel)
        self.engine.ylabel(ylabel)

        if logx:
            self.engine.xscale('log')
        if logy:
            self.engine.yscale('log')

        self.engine.plot(x, y, label=datalabel)


class MatrixPlotAdapter(PlotAdapter):

    def nuclidechart(self, matrix, xlabel="", ylabel="",
                     nstep=2, zstep=2, minX=0,  maxX=288,
                     minY=0, maxY=118, colourmap=cm.cool):
        self.newcanvas()
        self.grid()

        ax = self.engine.gca()
        r, c = matrix.shape

        ax.set_aspect('equal', 'box')
        boxsize = 1
        for (x, y), w in np.ndenumerate(matrix):
            if w <= 0.0:
                continue
            rect = Rectangle([x - boxsize / 2, y - boxsize / 2],
                             boxsize, boxsize,
                             facecolor=colourmap(w),
                             edgecolor='black')
            ax.add_patch(rect)

        if xlabel:
            ax.set_xlabel(xlabel)
        if ylabel:
            ax.set_ylabel(ylabel)

        ax.xaxis.set_ticks(np.arange(0, r+1, nstep))
        ax.yaxis.set_ticks(np.arange(0, c+1, zstep))
        ax.set_xlim((max(minX-1,0), maxX+1))
        ax.set_ylim((max(minY-1,0), maxY+1))
        ax.autoscale_view()

        self.engine.tight_layout()

    def matrixplot(self, matrix):
        self.newcanvas()
        self.grid()
        self.engine.imshow(matrix.T, origin='lower',
                           norm=LogNorm(vmin=10e-24, vmax=1),
                           cmap=cm.cool)
        self.engine.colorbar()

class AnimatedMatrixPlotAdapter(PlotAdapter):

    def animatedchart(self, matricies, xlabel="", ylabel="",
                      nstep=2, zstep=2, minX=0, maxX=100,
                      minY=0, maxY=100, colourmap=cm.cool,
                      timeinterval=100, figuresize=((15,10))):
        fig = self.newcanvas(figsize=figuresize)
        ax = self.engine.gca()
        self.grid()

        ax.xaxis.set_ticks(np.arange(0, maxX+1, nstep))
        ax.yaxis.set_ticks(np.arange(0, maxY+1, zstep))
        ax.set_xlim((max(minX-1,0), maxX+1))
        ax.set_ylim((max(minY-1,0), maxY+1))

        ax.set_aspect('equal', 'box')
        if xlabel:
            ax.set_xlabel(xlabel)
        if ylabel:
            ax.set_ylabel(ylabel)

        boxsize = 1
        def animate(i):
            [c.remove() for c in ax.collections]

            patches = []
            for (x, y), w in np.ndenumerate(matricies[i]):
                if w != 0.0:
                    patches.append(Rectangle([x - boxsize / 2, y - boxsize / 2],
                                     boxsize, boxsize,
                                     fc=colourmap(w),
                                     ec='black'))

            colors = np.logspace(10e-30, 1, 100, endpoint=True)
            p = PatchCollection(patches)
            p.set_edgecolor('black')
            p.set_array(np.array(colors))
            ax.add_collection(p)
            return ax,

        anim = animation.FuncAnimation(fig, animate,
                                       frames=len(matricies),
                                       interval=timeinterval,
                                       repeat_delay=timeinterval*3,
                                       blit=False)
        return anim



import os
import subprocess

from pypact.reader import Reader
from pypact.input.serialization import to_file
from pypact.util.file import file_remove, file_exists
from pypact.util.exceptions import PypactFispactExecutableNotFoundException
from pypact.util.decorators import time_it

FISPACT_EXE_PATH = os.getenv('FISPACT', os.path.join(os.sep, 'opt', 'fispact', 'bin', 'fispact'))


@time_it
def compute(input, files, fluxes,
            input_filename="fispacttemp.i",
            files_filename="fispacttemp.files",
            fluxes_filename="fispacttemp.fluxes",
            cleanup=True):

    files.fluxes = fluxes_filename
    runname = input_filename.strip('.i')
    
    # write all files
    to_file(input, input_filename)
    to_file(files, files_filename)
    to_file(fluxes, fluxes_filename)

    # run fispact
    if not file_exists(FISPACT_EXE_PATH):
        raise PypactFispactExecutableNotFoundException("Cannot find FISPACT-II binary. Instead got {}".format(FISPACT_EXE_PATH))

    print("* Running FISPACT-II...")
    proc = subprocess.check_call("{} {} {}".format(FISPACT_EXE_PATH, runname, files_filename), shell=True)
    print("* Computation complete")

    # check for log file
    logfile = "{}.log".format(runname)
    if not os.path.isfile(logfile):
        raise RuntimeError("No log file produced.")
    
    # clean up input and log files - keep output file
    if cleanup:
        file_remove(input_filename)
        file_remove(files_filename)
        file_remove(fluxes_filename)
        file_remove(logfile)

    outfile = "{}.out".format(runname)
    with Reader(outfile) as output:
        return output




from io import StringIO

def to_string(obj):
    stream = StringIO()
    obj._serialize(stream)
    s = stream.getvalue()
    stream.close()
    return s

def to_file(obj, filename):
    """
        Serialize data to test file (files file)
        
        obj: must have serialize method
        filename: the name of the file to write
    """
    with open(filename, 'w') as f:
        obj._serialize(f)


def from_file(obj, filename):
    """
        Deserialize data from file
        
        obj: must have deserialize and reset methods
        filename: the name of the file to read
    """
    obj.reset()
    with open(filename, 'r') as f:
        obj._deserialize(f)
